/* PC KEYBOARD Interface for AVR
 ***********************
 * DATA on PORTD BIT 5 *
 * SCK  on INTERRUPT 4 *
 ***********************
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "keyboard.h"

typedef unsigned char scan_code[2]; // for scancodes.h
#include "scancodes.h"

#define BUFF_SIZE 32

/* Datenleitung der Tastatur */
#define KEY_DATA (bit_is_set(PIND,5))

#define setbit(value,bit)   	( value =value|BV(bit) )
#define clearbit(value,bit) 	( value &= ~BV(bit) )
#define bitset(value,bit)   	( value & (1<<(bit)) )


volatile unsigned short  bitcount;
unsigned char kb_buffer[BUFF_SIZE];
unsigned char *inpt, *outpt;
volatile unsigned short  buffcnt;
unsigned char is_up=0,shift=0,funckey=0,altgr=0;
unsigned char keys_down[32];


void kbInit(void)  // external interupt Enable and stuff
{
 unsigned char i;
 inpt =  kb_buffer; // Initialize buffer
 outpt = kb_buffer;
 buffcnt = 0;
 bitcount = 11;
 
 for(i=0;i<32;i++) keys_down[i]=0;
 
 cbi(DDRD,5);   // Datenrichtung am Keyboard Data Bit, Eingang
 sbi(PORTD,5);  // Pull up 

 cbi(DDRE,4);
 sbi(PORTE,4);  // Pull up Wiederstand fr Interrrupt 4 einschalten
 sbi(EIMSK,INT4);   // externen Interrupt4 aktivieren
 sbi(EICRB,ISC41);  // Interrupt bei steigender Flanke
 cbi(EICRB,ISC40);
 
 sei(); // enable interrupts
 shift=0;
}


void put_kbbuff(unsigned char c)
{
    if (buffcnt<BUFF_SIZE)                        // If buffer not full
    {
        *inpt=c;                                // Put character into buffer
        inpt++;                                    // Increment pointer
        buffcnt++;
        if (inpt >= (kb_buffer + BUFF_SIZE)) inpt = kb_buffer;
    }
}

void edit_key(unsigned char x,unsigned char state)
{
	if(state) setbit(keys_down[x>>3],x & 7);
		else clearbit(keys_down[x>>3],x & 7);
}


void decode(unsigned char sc){
    unsigned char i;
// output scancode for debug !?
//    UART_PRINT("0x");
//    UART_Printfu08(sc);
//    UART_EOL();

    if (is_up)    {
        if(sc!=0xE0)
        {
        	is_up = 0;                         // Two 0xF0 in a row not allowed
       	}
        funckey=0;
    	edit_key(sc,0);
        switch (sc)
        {
          case 0x12 : shift = 0; break;                      // Left SHIFT
          case 0x59 : shift = 0; break;                      // Right SHIFT
          case 0x11 : altgr = 0; break;
    //      case 0x76 : break;                      // ESC
        }
    }else
    {
        edit_key(sc,1);
        switch (sc)
        {
          case 0xF0 : is_up = 1; break;      // The up-key identifier
          case 0x12 : shift = 1; break;      // Left SHIFT
          case 0x59 : shift = 1; break;      // Right SHIFT
          case 0x58 : shift = 1; break;      // Caps Lock
          case 0xE0 : funckey=1; break;
          case 0x11 : altgr = 1; break;
          
          default:
          { 
            if (funckey)
            {
            	put_kbbuff(0);
            	put_kbbuff(sc);
            	funckey=0;
            } else if (altgr)
            {
                for(i = 0; PRG_RDB(&alt_gr[i][0])!=sc && PRG_RDB(&alt_gr[i][0]); i++);
                if (PRG_RDB(&alt_gr[i][0]) == sc) {
                    put_kbbuff(PRG_RDB(&alt_gr[i][1]));
                }
            }
            else if(!shift)                    // If shift not pressed,
            {                             // do a table look-up
                for(i = 0; PRG_RDB(&unshifted[i][0])!=sc && PRG_RDB(&unshifted[i][0]); i++);
                if (PRG_RDB(&unshifted[i][0]) == sc) {
                    put_kbbuff(PRG_RDB(&unshifted[i][1]));
                }  else {put_kbbuff(0); put_kbbuff(sc);}
            } else {                    // If shift pressed
                for(i = 0; PRG_RDB(&shifted[i][0])!=sc && PRG_RDB(&shifted[i][0]); i++);
                if (PRG_RDB(&shifted[i][0]) == sc) {
                    put_kbbuff(PRG_RDB(&shifted[i][1]));
                }  else {put_kbbuff(0); put_kbbuff(sc);}
            }
          }break;  
        } 
        
    }
    
}

unsigned char  kbKeyPressed(void)	// returns true if a key is pressed
{
  return (buffcnt!=0);
}

unsigned char  kbReadKey(void)	// returns Ascii code of pressed key, waits until a key is pressed
{
    unsigned char byte;

    while(buffcnt==0);                        // Wait for data

    byte = *outpt;                                // Get byte
    outpt++;                                    // Increment pointer
    if (outpt >= (kb_buffer + BUFF_SIZE)) outpt = kb_buffer;
    buffcnt--;                                    // Decrement buffer count
    return byte;
}


unsigned char  kbKey(unsigned char x)		// returns true if Key x is pressed (scancode)
{
  return bitset(keys_down[x>>3],x & 7);
}

/* PC tastatur takt signal */
inline void kbInt(void){
    static unsigned char data;                // Holds the received scan code

    if ((bitcount >= 11) && KEY_DATA) return;  // kein Startbit, Startbit ist immer 0
    if ((bitcount <  11) && (bitcount > 2))    // Bit 3 to 10 is data. Parity bit is ignored
    {
      	data = (data >> 1);
       	if(KEY_DATA) data = data | 0x80;
    }
      
    if((--bitcount) == 0)                    // All bits received
    {	    // check stop bit, Stopbit ist immer 1
            if(KEY_DATA) decode(data);		
            data=0;
            bitcount = 11;
    }
}
